"""
The module contains functions for retrieving IAU meteor shower information.
"""
from typing import Dict

import requests

IAU_SHOWERS_LIST_URL = "https://www.ta3.sk/IAUC22DB/MDC2007/Etc/streamfulldata.txt"
"""The url that contains the list of IAU shower information."""


def get_iau_showers() -> Dict[str, Dict[str, str]]:
    """
    Gets the official list of IAU shower numbers, codes and names.

    :return: A dictionary, where the key is the shower number, of dictionaries
     containing the IAU shower information.
    :raises: requests.HTTPError if the source server doesn't return a 200 response.
    """
    response = requests.get(IAU_SHOWERS_LIST_URL)
    if not response.ok:
        response.raise_for_status()
        return {}  # pragma: no cover

    showers = {}
    for row in response.text.splitlines():
        if not row or row.startswith(":") or row.startswith("+"):
            continue

        row_values = row.split("|")

        shower_no = row_values[1].strip('" ')
        if shower_no in showers:
            continue

        shower_code = row_values[3].strip('" ')
        shower_name = row_values[4].strip('" ')

        showers[shower_no] = {"id": shower_no, "code": shower_code, "name": shower_name}

    return showers



"""
This module contains functions to load meteor/trajectory summary data into Pandas
DataFrames and NumPy arrays.
"""
import math
import os.path
from io import StringIO
from pathlib import Path
from typing import Any
from typing import List
from typing import Optional
from typing import Union

import numpy.typing as npt
import pandas as pd  # type: ignore
from pandas._typing import FilePathOrBuffer  # type: ignore

import gmn_python_api.meteor_summary_schema

"""The format of dates in meteor/trajectory summary data."""
DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S.%f"


def read_meteor_summary_csv_as_dataframe(
    filepath_or_buffer: Union[FilePathOrBuffer, List[FilePathOrBuffer]],
    camel_case_column_names: Optional[bool] = False,
    avro_compatible: Optional[bool] = False,
    avro_long_beginning_utc_time: Optional[bool] = True,
    csv_data_directory_format: Optional[bool] = False,
) -> pd.DataFrame:
    """
    Reads meteor summary REST API or trajectory summary data directory CSV data into a
     Pandas DataFrame. Columns available in the DataFrame can be found here:
     https://gmn-python-api.readthedocs.io/en/latest/data_schemas.html

    :param filepath_or_buffer: Path or buffer for a trajectory/meteor summary file. Or a
     list of paths or buffers (types can be mixed) that will be joined in the DataFrame.
     If using a list the first item must contain the csv header, and
     all items must either be all data directory CSVs or all REST API CSVs.
    :param camel_case_column_names: If True, column names will be camel cased e.g. m_deg
    :param avro_compatible: If True, the types and column names in the DataFrame will
     match exactly the AVSC schema with row.to_dict(). AVRO AVSC Schema can be found
     with the get_meteor_summary_avro_schema function.
    :param avro_long_beginning_utc_time: If True, the time column will be converted from
     a datetime object to an int64 epoch time which is compatible with the long
     timestamp-micros avro type.
    :param csv_data_directory_format: If True, the filepath_or_buffer headers will be
     treated as a CSV from the GMN Data Directory. If False, the filepath_or_buffer
     headers will be treated as a REST API CSV.

    :return: Pandas DataFrame of the meteor/trajectory summary data.
    :raises: TypeError if an invalid filepath_or_buffer type is provided.
    """
    joined_data = _join_filepath_or_buffer(filepath_or_buffer)

    if csv_data_directory_format:
        meteor_summary_df = pd.read_csv(
            StringIO(joined_data, newline="\n"),
            engine="python",
            sep=r"\s*;\s*",
            skiprows=[0, 5, 6],
            header=[0, 1],
            na_values=["nan", "...", "None"],
        )
        # Clean header text
        meteor_summary_df.columns = meteor_summary_df.columns.map(
            lambda h: f"{_clean_header(h[0])}{_clean_header(h[1], is_unit=True)}"
        )
    else:
        meteor_summary_df = pd.read_csv(
            StringIO(joined_data, newline="\r"), engine="python"
        )
        # Convert camel case column names to verbose names
        bidict = (
            gmn_python_api.meteor_summary_schema.get_verbose_and_camel_case_column_name_bidict()
        )
        for column in meteor_summary_df.columns:
            meteor_summary_df.rename(columns={column: bidict[column]}, inplace=True)

    # Set data types
    meteor_summary_df["Beginning (UTC Time)"] = pd.to_datetime(
        meteor_summary_df["Beginning (UTC Time)"], format=DATETIME_FORMAT
    )
    meteor_summary_df["IAU (code)"] = meteor_summary_df["IAU (code)"].astype("string")
    meteor_summary_df["IAU (No)"] = (
        meteor_summary_df["IAU (No)"].fillna(-1).astype("int64")
    )
    meteor_summary_df["Beg in (FOV)"] = meteor_summary_df["Beg in (FOV)"].map(
        {"True": True, "False": False}
    )
    meteor_summary_df["Beg in (FOV)"] = meteor_summary_df["Beg in (FOV)"].astype("bool")
    meteor_summary_df["End in (FOV)"] = meteor_summary_df["End in (FOV)"].map(
        {"True": True, "False": False}
    )
    meteor_summary_df["End in (FOV)"] = meteor_summary_df["End in (FOV)"].astype("bool")
    meteor_summary_df["Participating (stations)"] = meteor_summary_df[
        "Participating (stations)"
    ].astype("string")
    meteor_summary_df["Participating (stations)"] = meteor_summary_df[
        "Participating (stations)"
    ].apply(lambda x: x[1:-1].split(","))

    meteor_summary_df[
        "Schema (version)"
    ] = gmn_python_api.meteor_summary_schema.SCHEMA_VERSION
    meteor_summary_df["Schema (version)"] = meteor_summary_df[
        "Schema (version)"
    ].astype("string")

    meteor_summary_df.set_index("Unique trajectory (identifier)", inplace=True)

    if avro_compatible:
        camel_case_column_names = True

        if avro_long_beginning_utc_time:
            # convert datetime nano to micro epoch and round to int
            meteor_summary_df["Beginning (UTC Time)"] = (
                meteor_summary_df["Beginning (UTC Time)"].astype("int64") / 1e3
            )
            meteor_summary_df["Beginning (UTC Time)"] = (
                meteor_summary_df["Beginning (UTC Time)"].round(0).astype("int64")
            )

        meteor_summary_df["IAU (code)"] = meteor_summary_df["IAU (code)"].astype(
            "unicode"
        )
        meteor_summary_df["Schema (version)"] = meteor_summary_df[
            "Schema (version)"
        ].astype("unicode")

        # Convert null values to avro compatible types
        meteor_summary_df = meteor_summary_df.applymap(
            lambda x: None
            if x == "<NA>" or (isinstance(x, float) and math.isnan(x))
            else x
        )
        meteor_summary_df.reset_index(inplace=True)

    if camel_case_column_names:
        meteor_summary_df.columns = meteor_summary_df.columns.str.replace(
            "[^0-9a-zA-Z]+", "_", regex=True
        )
        meteor_summary_df.columns = meteor_summary_df.columns.str.rstrip("_")
        meteor_summary_df.columns = meteor_summary_df.columns.str.lstrip("_")
        meteor_summary_df.columns = meteor_summary_df.columns.str.replace(
            "Q_AU", "q_au_"
        )
        meteor_summary_df.columns = meteor_summary_df.columns.str.lower()
        meteor_summary_df.index.name = "unique_trajectory_identifier"

    return meteor_summary_df


def read_meteor_summary_csv_as_numpy_array(
    filepath_or_buffer: FilePathOrBuffer,
    camel_case_column_names: Optional[bool] = False,
    avro_compatible: Optional[bool] = False,
    avro_long_beginning_utc_time: Optional[bool] = True,
    csv_data_directory_format: Optional[bool] = False,
) -> npt.NDArray[Any]:
    """
    Reads meteor summary REST API or trajectory summary data directory CSV data into a
     NumPy array. Similar to read_meteor_summary_csv_as_dataframe. Columns available in
     the DataFrame can be found here:
     https://gmn-python-api.readthedocs.io/en/latest/data_schemas.html

    :param filepath_or_buffer: Path or buffer for a trajectory/meteor summary file. Or a
     list of paths or buffers (types can be mixed) that will be joined in the array.
     If using a list the first item must contain the csv header, and
     all items must either be all Data Directory CSVs or all REST API CSVs.
    :param camel_case_column_names: If True, column names will be camel cased e.g. m_deg
    :param avro_compatible: If True, the types and column names in the array will
     match exactly the AVSC schema with row.to_dict(). AVRO AVSC Schema can be found
     with the get_meteor_summary_avro_schema function.
    :param avro_long_beginning_utc_time: If True, the time column will be converted from
     a datetime object to an int64 epoch time which is compatible with the long
     timestamp-micros avro type.
    :param csv_data_directory_format: If True, the filepath_or_buffer headers will be
     treated as a CSV from the GMN Data Directory. If False, the filepath_or_buffer
     headers will be treated as a REST API CSV.

    :return: NumPy array of the meteor/trajectory summary data.
    :raises: TypeError if an invalid filepath_or_buffer type is provided.
    """
    data_frame = read_meteor_summary_csv_as_dataframe(
        filepath_or_buffer,
        camel_case_column_names=camel_case_column_names,
        avro_compatible=avro_compatible,
        avro_long_beginning_utc_time=avro_long_beginning_utc_time,
        csv_data_directory_format=csv_data_directory_format,
    )
    return data_frame.to_numpy()  # type: ignore


def _join_filepath_or_buffer(  # noqa: C901
    filepath_or_buffer: Union[FilePathOrBuffer, List[FilePathOrBuffer]]
) -> str:
    """
    Join data provided in the reader filepath_or_buffer parameter into a single CSV
     string that can be read by the reader function.

    :param filepath_or_buffer: Path or buffer for a trajectory/meteor summary file. Or a
     list of paths or buffers (types can be mixed) that will be joined in the dataframe.
     If using a list the first item must contain the csv header, and
     all items must either be all data directory CSVs or all REST API CSVs.
    :param csv_data_directory_format: If True, the filepath_or_buffer headers will be
     treated as a CSV from the GMN Data Directory. If False, the filepath_or_buffer
     headers will be treated as a REST API CSV.

    :return: The joined CSV data.
    :raises: TypeError if an invalid filepath_or_buffer type is provided.
    """
    if not isinstance(filepath_or_buffer, list):
        filepath_or_buffer_list = [filepath_or_buffer]
    else:
        filepath_or_buffer_list = filepath_or_buffer

    # Join list of data passed in to a single CSV
    joined_data = ""
    for i, item in enumerate(filepath_or_buffer_list):
        if type(item) is str:
            if os.path.isfile(item):
                new_data = open(item).read() + "\n"
            else:
                new_data = item + "\n"
        elif isinstance(item, Path):
            new_data = open(item).read() + "\n"
        else:
            raise TypeError(
                f"filepath_or_buffer must be of type string or bytes, or a list of those"
                f" types. Got {type(item)}."
            )

        # Remove multiple headers in data by only keeping the first header
        if i > 0:
            new_data_lines = new_data.split("\n")

            # For REST API CSV header
            if new_data_lines[0].startswith("unique_trajectory_identifier,"):
                new_data_lines = new_data_lines[1:]

            # For data directory CSV header
            # Splice out lines that start with a #, break after # stops
            elif new_data_lines[0][0] == "#":
                for j, line in enumerate(new_data_lines):  # pragma: no branch
                    line = line.replace("\n", "").replace("\r", "")
                    if line != "" and line[0] != "#":
                        new_data_lines = new_data_lines[j:]
                        break

            new_data = "\n".join(new_data_lines)

        joined_data += new_data
        i += 1

    return joined_data


def _clean_header(text: str, is_unit: bool = False) -> str:
    """
    Extract header text from each raw trajectory summary csv file header.

    :param text: Raw trajectory summary csv column header text.
    :param is_unit: If True, return text with brackets for units.

    :returns: Formatted text.
    """
    # Return an empty string if there is no header found
    if "Unnamed" in text:
        return ""

    # Removes additional spaces and hashtags from text. Add brackets optionally.
    clean_header = " ".join(text.replace("#", "").split())
    if is_unit:
        clean_header = f" ({clean_header})"

    return clean_header



"""
This module contains functions to read meteor summary data from the GMN REST API.
 The REST API uses the Datasette API endpoint. More info:
 https://docs.datasette.io/en/stable/json_api.html
"""
from enum import Enum
from typing import Dict
from typing import Optional
from typing import Tuple
from urllib.parse import urlencode

import requests

QUERY_URL = (
    "https://globalmeteornetwork.org/gmn_data_store"
    "{table}.{data_format}?_shape={data_shape}"
)
"""
URL template of GMN REST API endpoint. Note that the GMN REST API currently isn't live,
 so this URL should be replaced with a local instance of the GMN REST API to use
 functions in this module (e.g
 http:0.0.0.0:8001/gmn_data_store{table}.{data_format}?_shape={data_shape}).
"""


class DataFormat(str, Enum):
    """REST data format."""

    JSON = "json"
    CSV = "csv"


class DataShape(str, Enum):
    """REST data shape."""

    ARRAYS = "arrays"
    OBJECTS = "objects"
    ARRAY = "array"
    OBJECT = "object"


def get_meteor_summary_data_reader_compatible(
    where_sql: Optional[str] = None,
    next_page: Optional[str] = None,
) -> Tuple[str, Optional[str]]:
    """
    Get meteor summary data from the GMN REST API in a format that is compatible with
     the meteor_summary_reader functions.

    :param where_sql: An optional SQL WHERE clause to filter the data (e.g. iau_no='4').
    :param next_page: An optional URL to access more paginated data provided by the api.

    :return: The data returned from the GMN REST API in CSV format.
    :raises: requests.HTTPError If GMN REST API doesn't return a 200 response.
    """
    return get_meteor_summary_data(
        table_arguments={"_where": where_sql} if where_sql else None,
        data_format=DataFormat.CSV,
        data_shape=DataShape.ARRAY,
        next_page=next_page,
    )


def get_meteor_summary_data(
    table_arguments: Optional[Dict[str, str]] = None,
    data_format: DataFormat = DataFormat.JSON,
    data_shape: DataShape = DataShape.ARRAY,
    next_page: Optional[str] = None,
) -> Tuple[str, Optional[str]]:
    """
    Get meteor summary data from the GMN REST API.

    :param table_arguments: An optional dictionary of arguments to filter the data.
     A full list of arguments can be found here:
     https://docs.datasette.io/en/stable/json_api.html#table-arguments
    :param data_format: The data format of the meteor summary data using DataFormat
     enum.
    :param data_shape: The data shape of the meteor summary data using the DataShape
     enum.
    :param next_page: An optional URL to access more paginated data provided by the api.

    :return: The data returned from the GMN REST API.
    :raises: requests.HTTPError If GMN REST API doesn't return a 200 response.
    """
    return get_data(
        table="meteor_summary",
        table_arguments=table_arguments,
        data_format=data_format,
        data_shape=data_shape,
        next_page=next_page,
    )


def get_data_with_sql(
    sql: str,
    data_format: DataFormat = DataFormat.JSON,
    data_shape: DataShape = DataShape.ARRAY,
    next_page: Optional[str] = None,
) -> Tuple[str, Optional[str]]:
    """
    Get any available data from the GMN REST API using a readonly SQL query.

    :param sql: The SQL query to perform against the open access GMN Data Store database.
    :param data_format: The data format of the data using the DataFormat enum.
    :param data_shape: The data shape of the data using the DataShape enum.
    :param next_page: An optional URL to access more paginated data provided by the api.

    :return: The data returned from the GMN REST API.
    :raises: requests.HTTPError If GMN REST API doesn't return a 200 response.
    """
    return get_data(
        table_arguments={"sql": sql},
        data_format=data_format,
        data_shape=data_shape,
        next_page=next_page,
    )


def get_data(
    table: Optional[str] = None,
    table_arguments: Optional[Dict[str, str]] = None,
    data_format: DataFormat = DataFormat.JSON,
    data_shape: DataShape = DataShape.OBJECTS,
    next_page: Optional[str] = None,
) -> Tuple[str, Optional[str]]:
    """
    Get any data from the GMN REST API.

    :param table: The table to query in the GMN Data Store.
    :param table_arguments: An optional dictionary of arguments to filter the data. A
     full list of arguments can be found here:
     https://docs.datasette.io/en/stable/json_api.html#table-arguments
    :param data_format: The data format of the data using the DataFormat enum.
    :param data_shape: The data shape of the data using the DataShape enum.

    :return: The data returned from the GMN REST API.
    :raises: requests.HTTPError If GMN REST API doesn't return a 200 response.
    """
    if table_arguments is None:
        table_arguments = {}

    if next_page:
        query_url = next_page
    else:
        query_url = QUERY_URL.format(
            table="/" + table if table else None,
            data_format=data_format.value,
            data_shape=data_shape.value,
        )
        query_url += "&" + urlencode(table_arguments)

    data = _http_get_response(query_url)
    return data


def _http_get_response(url: str) -> Tuple[str, Optional[str]]:
    """
    Get resultant data from a given GMN REST API URL.

    :param url: A query URL of the GMN REST API.

    :return: Data, and a URL to access more paginated data if required.
    :raises: requests.HTTPError If URL doesn't return a 200 response.
    """
    response = requests.get(url)
    if response.ok:
        try:
            next_url = response.links.get("next").get("url")  # type: ignore
        except AttributeError:
            # next_url somtimes is not given even when paginated (for example with CSVs)
            if str(response.text).count("\n") == 101 and (
                requests.head(url + "&_next=100").ok
                or requests.head(url + "?_next=100").ok
            ):  # pragma: no cover
                next_url = url + "&_next=100" if "?" in url else url + "?_next=100"
            else:
                next_url = None
        return str(response.text), next_url
    else:
        response.raise_for_status()
        return "", ""  # pragma: no cover



"""GMN Python API."""
from gmn_python_api.data_directory import *  # noqa: F403
from gmn_python_api.gmn_rest_api import *  # noqa: F403
from gmn_python_api.iau_showers import *  # noqa: F403
from gmn_python_api.meteor_summary_reader import *  # noqa: F403
from gmn_python_api.meteor_summary_schema import *  # noqa: F403

__all__ = [  # noqa: F405
    "data_directory",
    "meteor_summary_reader",
    "iau_showers",
    "meteor_summary_schema",
    "gmn_rest_api",
]



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gmn_python_api']

package_data = \
{'': ['*'], 'gmn_python_api': ['data_models/*']}

install_requires = \
['avro-validator==1.0.9',
 'avro==1.11.0',
 'beautifulsoup4>=4.10.0,<5.0.0',
 'click==8.0.4',
 'numpy>1.20.3',
 'pandas>=1.1.0,<=1.3.5',
 'pandavro==1.6.0',
 'requests>=2.21.0,<3.0.0',
 'types-requests>=2.27.8,<3.0.0']

entry_points = \
{'console_scripts': ['gmn-python-api = gmn_python_api.__main__:main']}

setup_kwargs = {
    'name': 'gmn-python-api',
    'version': '0.0.6',
    'description': 'GMN Python API',
    'long_description': 'GMN Python API\n==============\n\n|PyPI| |Status| |Python Version| |License|\n\n|Read the Docs| |Tests| |Codecov|\n\n|pre-commit| |Black|\n\n.. |PyPI| image:: https://img.shields.io/pypi/v/gmn-python-api.svg\n   :target: https://pypi.org/project/gmn-python-api/\n   :alt: PyPI\n.. |Status| image:: https://img.shields.io/pypi/status/gmn-python-api.svg\n   :target: https://pypi.org/project/gmn-python-api/\n   :alt: Status\n.. |Python Version| image:: https://img.shields.io/pypi/pyversions/gmn-python-api\n   :target: https://pypi.org/project/gmn-python-api\n   :alt: Python Version\n.. |License| image:: https://img.shields.io/github/license/gmn-data-platform/gmn-python-api\n   :target: https://opensource.org/licenses/MIT\n   :alt: License\n.. |Read the Docs| image:: https://img.shields.io/readthedocs/gmn-python-api/latest.svg?label=Read%20the%20Docs\n   :target: https://gmn-python-api.readthedocs.io/\n   :alt: Read the documentation at https://gmn-python-api.readthedocs.io/\n.. |Tests| image:: https://github.com/gmn-data-platform/gmn-python-api/workflows/Tests/badge.svg\n   :target: https://github.com/gmn-data-platform/gmn-python-api/actions?query=workflow%3ATests+branch%3Amain\n   :alt: Tests\n.. |Codecov| image:: https://codecov.io/gh/gmn-data-platform/gmn-python-api/branch/main/graph/badge.svg\n   :target: https://codecov.io/gh/gmn-data-platform/gmn-python-api\n   :alt: Codecov\n.. |pre-commit| image:: https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white\n   :target: https://github.com/pre-commit/pre-commit\n   :alt: pre-commit\n.. |Black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n   :target: https://github.com/psf/black\n   :alt: Black\n\nPython package for accessing open `Global Meteor Network`_ (GMN) meteor trajectory `data`_.\nGlobal meteor data is generated using a network of low-light cameras pointed towards the night sky.\nMeteor properties (radiants, orbits, magnitudes and masses) are produced by the GMN and are available through this package.\n\n.. image:: https://raw.githubusercontent.com/gmn-data-platform/gmn-python-api/main/screenshot.png\n  :alt: Data screenshot\n\n|\n\n`Demo on Google Colab`_\n\nFeatures\n--------\n\n* Listing available daily and monthly csv trajectory summary files from the `GMN data directory`_.\n\n* Downloading specific daily and monthly csv trajectory summary files from the data directory.\n\n* Functions for loading the data directory trajectory summary data into Pandas_ DataFrames or Numpy_ arrays.\n\n* Functions for retrieving meteor summary data from the future GMN Data Store using the GMN REST API.\n\n* Functions for loading REST API meteor summary data into Pandas_ DataFrames or Numpy_ arrays.\n\n* Functions for retrieving the current meteor trajectory schema in AVRO_ format.\n\n* Functions for retrieving available IAU_ registered meteor showers.\n\nRequirements\n------------\n\n* Python 3.7.1+, 3.8, 3.9 or 3.10\n\n\nInstallation\n------------\n\nYou can install *GMN Python API* via pip_ from `PyPI`_:\n\n.. code:: console\n\n   $ pip install gmn-python-api\n\nOr install the latest development code, through TestPyPI_ or directly from GitHub_ via pip_:\n\n.. code:: console\n\n   $ pip install -i https://test.pypi.org/simple/ --extra-index-url https://pypi.org/simple gmn-python-api==<version>\n   Or\n   $ pip install git+https://github.com/gmn-data-platform/gmn-python-api\n\nThere is also a `development Google Colab notebook`_.\n\nSee the Troubleshooting_ section if you encounter installation issues.\n\nUsage\n-----\n\nSimple meteor analysis example:\n\n.. code:: python\n\n   from datetime import datetime\n   from gmn_python_api.data_directory import get_daily_file_content_by_date\n   from gmn_python_api.meteor_summary_reader import read_meteor_summary_csv_as_datafram\n\n   trajectory_summary_file_content = get_daily_file_content_by_date(datetime(2019, 7, 24))\n   trajectory_summary_dataframe = read_meteor_summary_csv_as_dataframe(\n       trajectory_summary_file_content,\n       csv_data_directory_format=True,\n   )\n\n   print(f"{trajectory_summary_dataframe[\'Vgeo (km/s)\'].max()} km/s "\n          "was the fastest geostationary velocity out of all meteors for that day.")\n   # 65.38499 km/s was the fastest geostationary velocity out of all meteors (24th of July 2019).\n\n   print(f"{trajectory_summary_dataframe.loc[trajectory_summary_dataframe[\'IAU (code)\'] == \'PER\'].shape[0]} "\n          "meteors were estimated to be part of the Perseids shower.")\n   # 8 meteors were estimated to be part of the Perseids shower (24th of July 2019).\n\n   print(f"Station {trajectory_summary_dataframe[\'Num (stat)\'].mode().values[0]} "\n          "recorded the highest number of meteors.")\n   # Station 2 recorded the highest number of meteors (24th of July 2019).\n\nPlease see the Usage_ and `API Reference`_ section for more details.\n\n\nContributing\n------------\n\nContributions are very welcome.\nTo learn more, see the `Contributor Guide`_.\n\n\nLicense\n-------\n\nDistributed under the terms of the `MIT license`_,\n*GMN Python API* is free and open source software.\n\n\nIssues\n------\n\nIf you encounter any problems,\nplease `file an issue`_ along with a detailed description.\n\n\nCredits\n-------\n\n`Hypermodern Python Cookiecutter`_ template.\n\n.. _Cookiecutter: https://github.com/audreyr/cookiecutter\n.. _MIT license: https://opensource.org/licenses/MIT\n.. _PyPI: https://pypi.org/project/gmn-python-api/\n.. _TestPyPI: https://test.pypi.org/project/gmn-python-api/\n.. _Hypermodern Python Cookiecutter: https://github.com/cjolowicz/cookiecutter-hypermodern-python\n.. _file an issue: https://github.com/gmn-data-platform/gmn-python-api/issues\n.. _pip: https://pip.pypa.io/\n.. github-only\n.. _Contributor Guide: https://gmn-python-api.readthedocs.io/en/latest/contributing.html\n.. _Usage: https://gmn-python-api.readthedocs.io/en/latest/usage.html\n.. _API Reference: https://gmn-python-api.readthedocs.io/en/latest/autoapi/gmn_python_api/index.html\n.. _Global Meteor Network: https://globalmeteornetwork.org/\n.. _data: https://globalmeteornetwork.org/data/\n.. _Demo on Google Colab: https://colab.research.google.com/github/gmn-data-platform/gmn-data-endpoints/blob/29f8c820bda950bf10d94e93c202857d73566e20/gmn_data_analysis_template.ipynb\n.. _GMN data directory: https://globalmeteornetwork.org/data/traj_summary_data/\n.. _Pandas: https://pandas.pydata.org/\n.. _Numpy: https://numpy.org/\n.. _GitHub: https://github.com/gmn-data-platform/gmn-python-api\n.. _Troubleshooting: https://gmn-python-api.readthedocs.io/en/latest/troubleshooting.html\n.. _development Google Colab notebook: https://colab.research.google.com/github/gmn-data-platform/gmn-data-endpoints/blob/cef0b3721737e8d65002d21dc56aa27d74003593/gmn_data_analysis_template_dev.ipynb\n.. _IAU: https://www.ta3.sk/IAUC22DB/MDC2007/\n.. _AVRO: https://avro.apache.org/docs/current/spec.html\n',
    'author': 'Ricky Bassom',
    'author_email': 'rickybas12@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/gmn-data-platform/gmn-python-api',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""
This module contains functions for handling the current meteor/trajectory summary data
 schema.
"""
import json
import os
import tempfile
from typing import Any
from typing import Dict

import pandavro as pdx  # type: ignore
from avro.datafile import DataFileReader  # type: ignore
from avro.io import DatumReader  # type: ignore

import gmn_python_api


SCHEMA_VERSION = "2.0"
"""The supported meteor/trajectory summary data format version."""

_MODEL_TRAJECTORY_SUMMARY_FILE_PATH = os.path.join(
    os.path.dirname(__file__),
    "data_models",
    "traj_summary_20220304_solrange_344.0-345.0.txt",
)
"""Model v2.0 trajectory summary file."""

_AVSC_PATH = os.path.join(
    tempfile.gettempdir(), f"trajectory_summary_schema_{SCHEMA_VERSION}.avsc"
)
"""The path for the temporary Avro schema file that will be returned as a string."""


def get_meteor_summary_avro_schema() -> Dict[str, Dict[str, Any]]:
    """
    Get the Avro schema (.avsc) for the current meteor/trajectory summary data format.

    :return: The Avro schema in .avsc format.
    """
    _, avro_file_path = tempfile.mkstemp()

    data_frame = gmn_python_api.read_meteor_summary_csv_as_dataframe(  # type: ignore
        _MODEL_TRAJECTORY_SUMMARY_FILE_PATH,
        avro_compatible=True,
        avro_long_beginning_utc_time=False,
        csv_data_directory_format=True,
    )

    pdx.to_avro(avro_file_path, data_frame)
    pdx.read_avro(avro_file_path)

    reader = DataFileReader(open(avro_file_path, "rb"), DatumReader())
    schema = json.loads(reader.meta["avro.schema"].decode())

    return dict(schema)


def get_verbose_and_camel_case_column_name_bidict() -> Dict[str, str]:
    """
    Get a bidirectional dictionary that maps the verbose and camel case column names.

    :return: A bidirectional dictionary that maps the verbose and camel case column
     names.
    """
    bidict = {}
    df_verbose = gmn_python_api.read_meteor_summary_csv_as_dataframe(  # type: ignore
        _MODEL_TRAJECTORY_SUMMARY_FILE_PATH,
        camel_case_column_names=False,
        csv_data_directory_format=True,
    )
    df_camel_case = gmn_python_api.read_meteor_summary_csv_as_dataframe(  # type: ignore
        _MODEL_TRAJECTORY_SUMMARY_FILE_PATH,
        camel_case_column_names=True,
        csv_data_directory_format=True,
    )

    bidict[df_verbose.index.name] = df_camel_case.index.name
    bidict[df_camel_case.index.name] = df_verbose.index.name

    for col in df_verbose.columns:
        bidict[col] = df_camel_case.columns[df_verbose.columns.get_loc(col)]
        bidict[df_camel_case.columns[df_verbose.columns.get_loc(col)]] = col

    return bidict



"""
This module contains functions to read trajectory summary files from the GMN data
directory.
"""
from datetime import datetime
from datetime import timedelta
from typing import List
from typing import Optional

import requests
from bs4 import BeautifulSoup  # type: ignore

BASE_URL: str = "https://globalmeteornetwork.org/data/traj_summary_data/"
"""The base URL for trajectory summary files in the GMN data directory."""

DATA_START_DATE: datetime = datetime(2018, 12, 9)
"""The date of the earliest trajectory summary file in the GMN data directory."""

DAILY_DIRECTORY: str = "daily/"
"""The name of the directory containing daily trajectory summary files in the
base URL."""

MONTHLY_DIRECTORY: str = "monthly/"
"""The name of the directory containing monthly trajectory summary files in
# the base URL."""

SUMMARY_FILE_EXTENSION: str = "txt"
"""The extension of the trajectory summary files in the GMN data directory"""

SUMMARY_TODAY_FILENAME: str = "traj_summary_latest_daily.txt"
"""The filename of the most recent trajectory summary file."""

SUMMARY_YESTERDAY_FILENAME: str = "traj_summary_yesterday.txt"
"""The filename of the trajectory summary file from yesterday."""

SUMMARY_ALL_FILENAME: str = "traj_summary_all.txt"
"""The filename of the trajectory summary file containing all data."""


def get_all_daily_file_urls() -> List[str]:
    """
    Get all daily trajectory summary file urls from the GMN data directory.

    :return: A list of all daily file urls.
    :raises: requests.HTTPError if the data directory url doesn't return a 200 response.
    """
    return _get_url_paths(BASE_URL + DAILY_DIRECTORY, SUMMARY_FILE_EXTENSION)


def get_all_monthly_file_urls() -> List[str]:
    """
    Get all monthly trajectory summary file urls from the GMN data directory.

    :return: A list of all monthly file urls.
    :raises: requests.HTTPError if the data directory url doesn't return a 200 response.
    """
    return _get_url_paths(BASE_URL + MONTHLY_DIRECTORY, SUMMARY_FILE_EXTENSION)


def get_all_file_url() -> str:
    """
    Get the URL of the trajectory summary file containing all data.

    :return: The URL of the file containing all data.
    """
    return BASE_URL + SUMMARY_ALL_FILENAME


def get_daily_file_url_by_date(
    date: datetime, current_date: Optional[datetime] = None
) -> str:
    """
    Get the URL of the daily trajectory summary file for a given date.

    :param date: The date of the daily file.
    :param current_date: The current date. Defaults to datetime.now().

    :return: The URL of the daily file.
    :raises: FileNotFoundError if the daily file cannot be found. Or
     requests.HTTPError is raised if the file url doesn't return a 200 response.
    """
    if not current_date:
        current_date = datetime.today()

    if date == current_date:
        return BASE_URL + DAILY_DIRECTORY + SUMMARY_TODAY_FILENAME

    if date == current_date - timedelta(days=1):
        return BASE_URL + DAILY_DIRECTORY + SUMMARY_YESTERDAY_FILENAME

    all_daily_filenames = get_all_daily_file_urls()
    files_containing_date = [
        f for f in all_daily_filenames if date.strftime("%Y%m%d") in f
    ]

    if len(files_containing_date) == 0:
        raise FileNotFoundError(f"Trajectory summary file not found for date {date}.")

    return files_containing_date[0]


def get_monthly_file_url_by_month(date: datetime) -> str:
    """
    Get the URL of the monthly trajectory summary file for a given month.

    :param date: The date of the monthly file.

    :return: The URL of the monthly file.
    :raises: FileNotFoundError if the monthly file cannot be found. Or
     requests.HTTPError is raised if the file url doesn't return a 200 response.
    """
    all_monthly_filenames = get_all_monthly_file_urls()
    files_containing_date = [
        f for f in all_monthly_filenames if date.strftime("%Y%m") in f
    ]

    if len(files_containing_date) == 0:
        raise FileNotFoundError(
            f"Trajectory summary file not found for month in date {date}."
        )

    return files_containing_date[0]


def get_file_content_from_url(file_url: str) -> str:
    """
    Get the content of a trajectory summary file from a given URL.

    :param url: The URL of the trajectory summary file.

    :return: The content of the file.
    :raises: requests.HTTPError If the file url doesn't return a 200 response.
    """
    response = requests.get(file_url)
    if response.ok:
        return str(response.text)
    else:
        response.raise_for_status()
        return ""  # pragma: no cover


def get_daily_file_content_by_date(
    date: datetime, current_date: Optional[datetime] = None
) -> str:
    """
    Get the content of the daily trajectory summary file for a given date.

    :param date: The date of the daily file.
    :param current_date: The current date. Defaults to datetime.now().

    :return: The content of the daily file.
    :raises: requests.HTTPError if the data directory url doesn't return a 200 response.
    """
    file_url = get_daily_file_url_by_date(date, current_date)
    return get_file_content_from_url(file_url)


def get_monthly_file_content_by_date(date: datetime) -> str:
    """
    Get the content of the monthly trajectory summary file for a given date.

    :param date: The date to get the monthly file for.

    :return: The content of the monthly file.
    :raises: requests.HTTPError if the data directory url doesn't return a 200 response.
    """
    file_url = get_monthly_file_url_by_month(date)
    return get_file_content_from_url(file_url)


def get_all_file_content() -> str:
    """
    Get the content of the trajectory summary file containing all data.

    :return: The content of the file containing all data.
    :raises: requests.HTTPError if the data directory url doesn't return a 200 response.
    """
    file_url = get_all_file_url()
    return get_file_content_from_url(file_url)


def _get_url_paths(url: str, ext: str = "") -> List[str]:
    """
    Get all paths from a directory listing URL.

    :param url: The URL to get the paths from.
    :param ext: The extension to filter by.

    :return: A list of all paths.
    :raises: requests.HTTPError if the URL doesn't return a 200 response.
    """
    response_text = get_file_content_from_url(url)
    soup = BeautifulSoup(response_text, "html.parser")
    parent = [
        url + node.get("href")
        for node in soup.find_all("a")
        if node.get("href").endswith(ext)
    ]
    return parent
